@@ -1,5 +1,5 @@
 # ext/declarative/api.py
-# Copyright (C) 2005-2019 the SQLAlchemy authors and contributors
+# Copyright (C) 2005-2021 the SQLAlchemy authors and contributors
 # <see AUTHORS file>
 #
 # This module is part of SQLAlchemy and is released under
@@ -14,14 +14,19 @@ from .base import _add_attribute
 from .base import _as_declarative
 from .base import _declarative_constructor
 from .base import _DeferredMapperConfig
+from .base import _del_attribute
+from .base import _get_immediate_cls_attr
 from .clsregistry import _class_resolver
 from ... import exc
+from ... import inspection
 from ... import util
 from ...orm import attributes
 from ...orm import comparable_property
+from ...orm import exc as orm_exc
 from ...orm import interfaces
 from ...orm import properties
 from ...orm import synonym as _orm_synonym
+from ...orm.base import _inspect_mapped_class
 from ...orm.base import _mapper_or_none
 from ...orm.util import polymorphic_union
 from ...schema import MetaData
@@ -66,7 +71,7 @@ def has_inherited_table(cls):
 
 
 class DeclarativeMeta(type):
-    def __init__(cls, classname, bases, dict_):
+    def __init__(cls, classname, bases, dict_, **kw):
         if "_decl_class_registry" not in cls.__dict__:
             _as_declarative(cls, classname, cls.__dict__)
         type.__init__(cls, classname, bases, dict_)
@@ -74,12 +79,16 @@ class DeclarativeMeta(type):
     def __setattr__(cls, key, value):
         _add_attribute(cls, key, value)
 
+    def __delattr__(cls, key):
+        _del_attribute(cls, key)
+
 
 def synonym_for(name, map_column=False):
-    """Decorator that produces an :func:`.orm.synonym` attribute in conjunction
+    """Decorator that produces an :func:`_orm.synonym`
+    attribute in conjunction
     with a Python descriptor.
 
-    The function being decorated is passed to :func:`.orm.synonym` as the
+    The function being decorated is passed to :func:`_orm.synonym` as the
     :paramref:`.orm.synonym.descriptor` parameter::
 
         class MyClass(Base):
@@ -101,7 +110,7 @@ def synonym_for(name, map_column=False):
 
         :ref:`synonyms` - Overview of synonyms
 
-        :func:`.orm.synonym` - the mapper-level function
+        :func:`_orm.synonym` - the mapper-level function
 
         :ref:`mapper_hybrids` - The Hybrid Attribute extension provides an
         updated approach to augmenting attribute behavior more flexibly than
@@ -184,12 +193,6 @@ class declared_attr(interfaces._MappedAttribute, property):
                 else:
                     return {"polymorphic_identity":cls.__name__}
 
-    .. versionchanged:: 0.8 :class:`.declared_attr` can be used with
-       non-ORM or extension attributes, such as user-defined attributes
-       or :func:`.association_proxy` objects, which will be assigned
-       to the class at class construction time.
-
-
     """
 
     def __init__(self, fget, cascading=False):
@@ -254,7 +257,8 @@ class declared_attr(interfaces._MappedAttribute, property):
                 def id(cls):
                     if has_inherited_table(cls):
                         return Column(
-                            ForeignKey('myclass.id'), primary_key=True)
+                            ForeignKey('myclass.id'), primary_key=True
+                        )
                     else:
                         return Column(Integer, primary_key=True)
 
@@ -348,7 +352,7 @@ def declarative_base(
 
     :param class_registry: optional dictionary that will serve as the
       registry of class names-> mapped classes when string names
-      are used to identify classes inside of :func:`.relationship`
+      are used to identify classes inside of :func:`_orm.relationship`
       and others.  Allows two or more declarative base classes
       to share the same registry of class names for simplified
       inter-base relationships.
@@ -413,8 +417,6 @@ def as_declarative(**kw):
     All keyword arguments passed to :func:`.as_declarative` are passed
     along to :func:`.declarative_base`.
 
-    .. versionadded:: 0.8.3
-
     .. seealso::
 
         :func:`.declarative_base`
@@ -463,24 +465,36 @@ class ConcreteBase(object):
                             'polymorphic_identity':'manager',
                             'concrete':True}
 
+
+    The name of the discriminator column used by :func:`.polymorphic_union`
+    defaults to the name ``type``.  To suit the use case of a mapping where an
+    actual column in a mapped table is already named ``type``, the
+    discriminator name can be configured by setting the
+    ``_concrete_discriminator_name`` attribute::
+
+        class Employee(ConcreteBase, Base):
+            _concrete_discriminator_name = '_concrete_discriminator'
+
+    .. versionadded:: 1.3.19 Added the ``_concrete_discriminator_name``
+       attribute to :class:`_declarative.ConcreteBase` so that the
+       virtual discriminator column name can be customized.
+
     .. seealso::
 
         :class:`.AbstractConcreteBase`
 
         :ref:`concrete_inheritance`
 
-        :ref:`inheritance_concrete_helpers`
-
 
     """
 
     @classmethod
-    def _create_polymorphic_union(cls, mappers):
+    def _create_polymorphic_union(cls, mappers, discriminator_name):
         return polymorphic_union(
             OrderedDict(
                 (mp.polymorphic_identity, mp.local_table) for mp in mappers
             ),
-            "type",
+            discriminator_name,
             "pjoin",
         )
 
@@ -490,10 +504,15 @@ class ConcreteBase(object):
         if m.with_polymorphic:
             return
 
+        discriminator_name = (
+            _get_immediate_cls_attr(cls, "_concrete_discriminator_name")
+            or "type"
+        )
+
         mappers = list(m.self_and_descendants)
-        pjoin = cls._create_polymorphic_union(mappers)
+        pjoin = cls._create_polymorphic_union(mappers, discriminator_name)
         m._set_with_polymorphic(("*", pjoin))
-        m._set_polymorphic_on(pjoin.c.type)
+        m._set_polymorphic_on(pjoin.c[discriminator_name])
 
 
 class AbstractConcreteBase(ConcreteBase):
@@ -511,6 +530,32 @@ class AbstractConcreteBase(ConcreteBase):
     and is only used for selecting.  Compare to :class:`.ConcreteBase`,
     which does create a persisted table for the base class.
 
+    .. note::
+
+        The :class:`.AbstractConcreteBase` class does not intend to set up  the
+        mapping for the base class until all the subclasses have been defined,
+        as it needs to create a mapping against a selectable that will include
+        all subclass tables.  In order to achieve this, it waits for the
+        **mapper configuration event** to occur, at which point it scans
+        through all the configured subclasses and sets up a mapping that will
+        query against all subclasses at once.
+
+        While this event is normally invoked automatically, in the case of
+        :class:`.AbstractConcreteBase`, it may be necessary to invoke it
+        explicitly after **all** subclass mappings are defined, if the first
+        operation is to be a query against this base class.  To do so, invoke
+        :func:`.configure_mappers` once all the desired classes have been
+        configured::
+
+            from sqlalchemy.orm import configure_mappers
+
+            configure_mappers()
+
+        .. seealso::
+
+            :func:`_orm.configure_mappers`
+
+
     Example::
 
         from sqlalchemy.ext.declarative import AbstractConcreteBase
@@ -528,11 +573,13 @@ class AbstractConcreteBase(ConcreteBase):
                 'polymorphic_identity':'manager',
                 'concrete':True}
 
+        configure_mappers()
+
     The abstract base class is handled by declarative in a special way;
     at class configuration time, it behaves like a declarative mixin
     or an ``__abstract__`` base class.   Once classes are configured
     and mappings are produced, it then gets mapped itself, but
-    after all of its decscendants.  This is a very unique system of mapping
+    after all of its descendants.  This is a very unique system of mapping
     not found in any other SQLAlchemy system.
 
     Using this approach, we can specify columns and properties
@@ -564,6 +611,8 @@ class AbstractConcreteBase(ConcreteBase):
                 'polymorphic_identity':'manager',
                 'concrete':True}
 
+        configure_mappers()
+
     When we make use of our mappings however, both ``Manager`` and
     ``Employee`` will have an independently usable ``.company`` attribute::
 
@@ -579,8 +628,6 @@ class AbstractConcreteBase(ConcreteBase):
 
         :ref:`concrete_inheritance`
 
-        :ref:`inheritance_concrete_helpers`
-
     """
 
     __no_table__ = True
@@ -608,7 +655,12 @@ class AbstractConcreteBase(ConcreteBase):
             mn = _mapper_or_none(klass)
             if mn is not None:
                 mappers.append(mn)
-        pjoin = cls._create_polymorphic_union(mappers)
+
+        discriminator_name = (
+            _get_immediate_cls_attr(cls, "_concrete_discriminator_name")
+            or "type"
+        )
+        pjoin = cls._create_polymorphic_union(mappers, discriminator_name)
 
         # For columns that were declared on the class, these
         # are normally ignored with the "__no_table__" mapping,
@@ -627,7 +679,7 @@ class AbstractConcreteBase(ConcreteBase):
 
         def mapper_args():
             args = m_args()
-            args["polymorphic_on"] = pjoin.c.type
+            args["polymorphic_on"] = pjoin.c[discriminator_name]
             return args
 
         to_map.mapper_args_fn = mapper_args
@@ -639,23 +691,35 @@ class AbstractConcreteBase(ConcreteBase):
             if sm and sm.concrete and cls in scls.__bases__:
                 sm._set_concrete_base(m)
 
+    @classmethod
+    def _sa_raise_deferred_config(cls):
+        raise orm_exc.UnmappedClassError(
+            cls,
+            msg="Class %s is a subclass of AbstractConcreteBase and "
+            "has a mapping pending until all subclasses are defined. "
+            "Call the sqlalchemy.orm.configure_mappers() function after "
+            "all subclasses have been defined to "
+            "complete the mapping of this class."
+            % orm_exc._safe_cls_name(cls),
+        )
+
 
 class DeferredReflection(object):
     """A helper class for construction of mappings based on
     a deferred reflection step.
 
     Normally, declarative can be used with reflection by
-    setting a :class:`.Table` object using autoload=True
+    setting a :class:`_schema.Table` object using autoload=True
     as the ``__table__`` attribute on a declarative class.
-    The caveat is that the :class:`.Table` must be fully
+    The caveat is that the :class:`_schema.Table` must be fully
     reflected, or at the very least have a primary key column,
     at the point at which a normal declarative mapping is
-    constructed, meaning the :class:`.Engine` must be available
+    constructed, meaning the :class:`_engine.Engine` must be available
     at class declaration time.
 
     The :class:`.DeferredReflection` mixin moves the construction
     of mappers to be at a later point, after a specific
-    method is called which first reflects all :class:`.Table`
+    method is called which first reflects all :class:`_schema.Table`
     objects created so far.   Classes can define it as such::
 
         from sqlalchemy.ext.declarative import declarative_base
@@ -705,13 +769,11 @@ class DeferredReflection(object):
         ReflectedOne.prepare(engine_one)
         ReflectedTwo.prepare(engine_two)
 
-    .. versionadded:: 0.8
-
     """
 
     @classmethod
     def prepare(cls, engine):
-        """Reflect all :class:`.Table` objects for all current
+        """Reflect all :class:`_schema.Table` objects for all current
         :class:`.DeferredReflection` subclasses"""
 
         to_map = _DeferredMapperConfig.classes_for_base(cls)
@@ -750,6 +812,16 @@ class DeferredReflection(object):
         if local_table is not None:
             cls._reflect_table(local_table, engine)
 
+    @classmethod
+    def _sa_raise_deferred_config(cls):
+        raise orm_exc.UnmappedClassError(
+            cls,
+            msg="Class %s is a subclass of DeferredReflection.  "
+            "Mappings are not produced until the .prepare() "
+            "method is called on the class hierarchy."
+            % orm_exc._safe_cls_name(cls),
+        )
+
     @classmethod
     def _reflect_table(cls, table, engine):
         Table(
@@ -761,3 +833,17 @@ class DeferredReflection(object):
             autoload_with=engine,
             schema=table.schema,
         )
+
+
+@inspection._inspects(DeclarativeMeta)
+def _inspect_decl_meta(cls):
+    mp = _inspect_mapped_class(cls)
+    if mp is None:
+        if _DeferredMapperConfig.has_cls(cls):
+            _DeferredMapperConfig.raise_unmapped_for_cls(cls)
+            raise orm_exc.UnmappedClassError(
+                cls,
+                msg="Class %s has a deferred mapping on it.  It is not yet "
+                "usable as a mapped class." % orm_exc._safe_cls_name(cls),
+            )
+    return mp
